import os
import sys
import re
import itertools


class TE(object):

    def __init__(self, chr, id):
        self.chr = chr
        self.id = id
        self.nester_domains = []
        self.dante_domains = {}
        self.ltr_loc = []
        self.full_id = "{}|{}".format(chr, id)

    def final_anot(self):
        return max(self.dante_domains, key=lambda x: self.dante_domains[x][0])

    def max_score(self):
        return self.dante_domains[max(self.dante_domains, key=lambda x: self.dante_domains[x][0])][0]


class Domain(object):

    def __init__(self, name, loc):
        self.name = name
        self.loc = loc


def parse_nester(fileName):
    elements = []
    processed = []
    current_element = None
    te_base_pattern = r"(?<=ID=TE_BASE_)\d+"
    with open(fileName) as f:
        for line in f:
            if line[0] == "#":
                continue
            tmp = line.rstrip().split("\t")
            if "nested_repeat" in line:
                el_num = re.search(te_base_pattern, line).group(0)
                idd = tmp[0] + "|" + el_num
                if idd not in processed:
                    processed.append(idd)
                    current_element = TE(tmp[0], el_num)
                    elements.append(current_element)
                continue
            elif "polypeptide_conserved_region" in line:
                attributes = tmp[8].split(";")
                loc = [int(tmp[3]), int(tmp[4])]
                name = attributes[-1].split("=")[1]
                current_element.nester_domains.append(Domain(name, loc))

            elif "long_terminal_repeat" in line:
                current_element.ltr_loc.append([int(tmp[3]), int(tmp[4])])
    return elements


def parse_dante(dante, nester):
    elements = parse_nester(nester)
    current_id = ""
    current_element = None
    processed = []
    with open(dante) as f:
        for line in itertools.islice(f, 5, None):
            tmp = line.rstrip().split("\t")
            attributes = tmp[8].split(";")
            idd = "{}|{}".format(tmp[0], attributes[-1].split("-")[1])
            if idd in processed:
                continue            
            if current_id == "":
                current_id = idd
                current_element = fetch_element(elements, current_id)
            if current_id == idd:
                if current_element is None:
                    continue
                add_dante_domain(current_element, attributes, [int(tmp[3]), int(tmp[4])])
            else:
                processed.append(current_id)
                current_id = idd
                current_element = fetch_element(elements, current_id)
                if current_element is None:
                    continue
                add_dante_domain(current_element, attributes, [int(tmp[3]), int(tmp[4])])
    return elements


def add_dante_domain(element, attributes, loc):
    family = attributes[1].split("=")[1]
    name = attributes[0].split("=")[1]
    if family not in element.dante_domains.keys():
        element.dante_domains[family] = [0, Domain(name, loc)]
    else:
        element.dante_domains[family].append(Domain(name, loc))


def fetch_element(elements, idd):
    for el in elements:
        if el.full_id == idd:
            return el
    return None


def determine_family(dante, nester):
    elements = parse_dante(dante, nester)    
    dante_to_nester = { "PROT" : "AP", "RH" : "RNaseH", "RT" : "RT", "GAG" : "GAG", "INT" : "INT" , "CHDII" : "", "aRH" : "", "CHDCR" : "", "ENDO" : "", "TPase" : "", "HEL2" : ""}
    for el in elements:
        for nester_domain in el.nester_domains:
            for key in el.dante_domains.keys():
                count = 0
                for dante_domain in el.dante_domains[key][1:]:
                    if dante_domain.name not in dante_to_nester.keys():
                        continue
                    if nester_domain.name == dante_to_nester[dante_domain.name]:
                        count += intersect(nester_domain.loc, dante_domain.loc)
                el.dante_domains[key][0] += count
    return elements


def get_output(dante_origin, dante_nested, nester):
    elements_origin = determine_family(dante_origin, nester)
    elements_nested = determine_family(dante_nested, nester)
    output = open("{}_annot.gff".format(os.path.splitext(nester)[0]), 'w')
    bordel = open("{}_trash.gff".format(os.path.splitext(nester)[0]), 'w')
    tsrless = open("{}_no_tsr.gff".format(os.path.splitext(nester)[0]), 'w') #multiple dante domains in ltr
    pair = []
    multiple_domains_in_ltr = 0
    with open(nester) as f:
        for line in itertools.islice(f, 1, None):
            if line[0] == "#":
                proc_pair, both, orig_tsr = process_pair(pair, elements_origin, elements_nested)
                if both:
                    if orig_tsr:
                        if not domains_in_ltr(pair, elements_origin, elements_nested):
                            output.write(line + proc_pair)
                        else:
                            tsrless.write(line + proc_pair)
                    else:
                        tsrless.write(line + proc_pair)
                else:
                    bordel.write(line + proc_pair)
                pair = []
            else:
                pair.append(line.rstrip())


def process_pair(pair, origin, nested):
    d = 0
    for i in range(len(pair)):
        if "nested_repeat" in pair[i]:
            p = False
            for el in origin:
                if el.chr in pair[i] and "ID=TE_BASE_{}".format(el.id) in pair[i] and correct_el(el):
                    pair[i] += ";annotation={};position={}".format(el.final_anot(), "nested" if d == 0 else "original")
                    d += 1
                    p = True
                    break
            if not p:
                for el in nested:
                    if el.chr in pair[i] and "ID=TE_BASE_{}".format(el.id) in pair[i] and correct_el(el):
                        pair[i] += ";annotation={};position={}".format(el.final_anot(), "nested" if d == 0 else "original")
                        d += 1
                        p = True
                        break
    tsr = orig_tsr(pair)
    return "\n".join(pair) + "\n", d == 2, tsr


def orig_tsr(pair):
    c = 0
    for line in pair:
        if "nested_repeat" in line:
            c += 1
        if c == 2 and "target_site_duplication" in line:
            return True
    return False

def domains_in_ltr(pair, elements_origin, elements_nested):
    te_base_pattern = r"(?<=ID=TE_BASE_)\d+"
    mult = 0
    for line in pair:
        if "nested_repeat" in line:
            fst = []
            snd = []
            tmp = line.split("\t")
            current_element = fetch_element(elements_origin, "{}|{}".format(tmp[0], re.search(te_base_pattern, line).group(0)))
            if current_element is None:
                current_element = fetch_element(elements_nested, "{}|{}".format(tmp[0], re.search(te_base_pattern, line).group(0))) 
            if current_element is not None:
                for key in current_element.dante_domains.keys():
                    for domain in current_element.dante_domains[key][1:]:
                        if intersect(current_element.ltr_loc[0], domain.loc):
                            fst.append(domain.loc[0])
                        if intersect(current_element.ltr_loc[1], domain.loc):
                            snd.append(domain.loc[0])
                if len(set(fst)) > 1 or len(set(snd)) > 1:
                    mult += 1
    return True if (mult > 0) else False
                
                                               
def correct_el(el):
    return (len(el.dante_domains.keys()) > 1 and el.dante_domains[el.final_anot()][0] > 19) or (len(el.dante_domains.keys()) == 1)


def intersect(nester, dante):
    if nester[0] <= dante[0] and nester[1] >= dante[0]:
        return min([nester[1], dante[1]]) - dante[0] + 1
    if nester[0] >= dante[0] and nester[0] <= dante[1]:
        return min([nester[1], dante[1]]) - nester[0] + 1
    return 0


get_output(sys.argv[1], sys.argv[2], sys.argv[3])
